
// local
#include "RichMissingHPDMonitor.h"

//-----------------------------------------------------------------------------

using namespace Rich::Mon;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
MissingHPDMonitor::MissingHPDMonitor( const std::string& name,
                                      ISvcLocator* pSvcLocator )
  : Rich::HistoAlgBase  ( name , pSvcLocator ),
    m_taeEvents         ( 1, ""              ),
    m_Name              ( name               )
{
  declareProperty( "RawEventLocations", m_taeEvents   );
  declareProperty( "HPDCheckFreq", m_checkFreq = 1000 );
  declareProperty( "MaxMissingEvents", m_maxMissingEvents = 20000 );
  declareProperty( "VetoNoBias"      , m_vetoNoBias       = true  );

  // Partition
  const char* partitionName = getenv("PARTITION");
  m_activeRICH[Rich::Rich1] = true;
  m_activeRICH[Rich::Rich2] = true;
  if ( partitionName )
  {
    const std::string sPartition(partitionName);
    m_Name += sPartition;
    if      ( "RICH1" == sPartition ) { m_activeRICH[Rich::Rich2] = false; }
    else if ( "RICH2" == sPartition ) { m_activeRICH[Rich::Rich1] = false; }
  }
}

//=============================================================================
// Initialization
//=============================================================================
StatusCode MissingHPDMonitor::initialize()
{
  const StatusCode sc = Rich::HistoAlgBase::initialize();
  if ( sc.isFailure() ) return sc;

  // talk to camera
  _ri_verbo << "send message to CAMERA" << endmsg;
  cameraTool()->Append("TEXT",m_Name.c_str());
  cameraTool()->SendAndClearTS(ICameraTool::INFO,m_Name,"Initialized");

  // return
  return sc;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode MissingHPDMonitor::execute()
{
  StatusCode sc = StatusCode::SUCCESS;

  // load the ODIN
  const auto * odin = getIfExists<LHCb::ODIN>( LHCb::ODINLocation::Default );
  if ( !odin ) return sc;

  // Veto NoBias
  if ( m_vetoNoBias )
  {
    const Gaudi::Math::Bit< unsigned int, 2 > bit{};
    if ( UNLIKELY( bit(odin->eventType()) ) ) 
    {
      //info() << "Rejecting No bias event" << endmsg;
      return sc; 
    }
    //else { info() << "Selecting event" << endmsg; }
  }

  // count events and print on the first
  if ( 1 == ++m_nEvts )
  {
    const std::string m = "First event seen";
    info() << m << endmsg;
    cameraTool()->Append("TEXT",m_Name.c_str());
    cameraTool()->SendAndClearTS(ICameraTool::INFO,m_Name,m);
    m_lastEventCheck = m_nEvts;
  }

  // main loop over HPDs in the data
  const auto & l1Map = smartIDDecoder()->allRichSmartIDs(m_taeEvents);
  _ri_verbo << "L1 map has size " << l1Map.size() << endmsg;

  for ( const auto & L1 : l1Map )
  {
    for ( const auto & ingress : L1.second )
    {
      for ( const auto & HPD : ingress.second.pdData() )
      {
        // HPD smart ID
        const auto & smartIDHPD = HPD.second.pdID();

        // check for valid HPD
        if ( !smartIDHPD.isValid() ) { continue; }

        // For testing, intentionally ignore an HPD
        //if ( smartIDHPD == LHCb::RichSmartID(Rich::Rich1,Rich::top,10,3) ) continue;

        // Flag this HPD as having been seen in the data, with the event count number
        m_hpdCount[ smartIDHPD ] = m_nEvts;
      }
    }
  }

  // Monitor things every now and then
  if ( m_nEvts                    >  m_maxMissingEvents &&
       (m_nEvts-m_lastEventCheck) >= m_checkFreq         )
  {

    // Loop over HPDs expected to be present and flag those missing
    for ( const auto & HPD : deRichSys()->activePDRichSmartIDs() )
    {
      // Is this RICH active.
      if ( m_activeRICH[HPD.rich()] )
      {
        if ( UNLIKELY( ( m_nEvts - m_hpdCount[HPD] ) > m_maxMissingEvents ) )
        {
          // it is missing
          std::ostringstream mess, hpdS;
          mess << "HPD active but no data seen for " << m_maxMissingEvents << " events or more";
          hpdDisableTool() -> DisableHPD( HPD, mess.str(), 10, m_nEvts-m_lastEventCheck );
          hpdS << HPD;
          Warning( hpdS.str()+" "+mess.str(), StatusCode::SUCCESS );
        }
      }
    }

    // Now, loop over those expected to be missing and flag those present
    for ( const auto & HPD : deRichSys()->inactivePDRichSmartIDs() )
    {
      // Is this RICH active.
      if ( m_activeRICH[HPD.rich()] )
      {
        if ( UNLIKELY( ( m_nEvts - m_hpdCount[HPD] ) < m_maxMissingEvents ) )
        {
          // This HPD is present in the data ...
          std::ostringstream mess;
          mess << "HPD inactive but present in the data during the last "
               << m_maxMissingEvents << " events";
          hpdDisableTool() -> ReportHPD( HPD, mess.str() );
          //warning() << HPD << " " << mess.str() << endmsg;
        }
      }
    }

    // set the last seen event number
    m_lastEventCheck = m_nEvts;

  }

  return sc;
}

//=============================================================================
//  Finalize
//=============================================================================
StatusCode MissingHPDMonitor::finalize()
{
  cameraTool()->SendAndClearTS(ICameraTool::INFO,m_Name,"Finalized");
  return Rich::HistoAlgBase::finalize();
}

//=============================================================================

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( MissingHPDMonitor )

//=============================================================================
