// $Id: $
// Include files

// from Gaudi
#include "GaudiKernel/AlgFactory.h"

// local
#include "RichTimingScanAnalysis.h"

//-----------------------------------------------------------------------------
// Implementation file for class : RichTimingScanAnalysis
//
// 2007-12-02 : Nicholas Styles
// 2009-10-02 : Conor Fitzpatrick
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( RichTimingScanAnalysis )

  //=============================================================================
  // Standard constructor, initializes variables
  //=============================================================================
RichTimingScanAnalysis::RichTimingScanAnalysis(const std::string& name,
                                               ISvcLocator* pSvcLocator)
: Rich::HistoAlgBase(name , pSvcLocator)
{
  declareProperty("NEventsPerStep", m_nEventsPerStep = 100000);
  declareProperty("NSteps", m_nSteps = 75);
  declareProperty("WhichRich", m_whichRich = 2);
  declareProperty("StepWidth", m_stepWidth= 2);
  declareProperty("StepStart", m_stepStart= 0);
  declareProperty("MinClusterSize", m_minClusSize = 1);
  declareProperty("MaxClusterSize", m_maxClusSize = 999999);
  declareProperty("CheckBxID" , m_filterBX = false );
  declareProperty("ValidBxIDs" , m_bxIDs = std::vector<long>{1001,27731});
}

//=============================================================================
// Destructor
//=============================================================================
RichTimingScanAnalysis::~RichTimingScanAnalysis() {}

//=============================================================================
// Initialization
//=============================================================================
bool RichTimingScanAnalysis::checkODINBxID()
{
  LHCb::ODIN* odin = get<LHCb::ODIN>( LHCb::ODINLocation::Default );
  debug() << "BxID: " << odin->bunchId() << endmsg;
  for (auto ibx = m_bxIDs.begin(); ibx != m_bxIDs.end(); ++ibx )
  {
    if ( (*ibx) ==  odin->bunchId() )
    {
      return true ;
    }
  }
  return false ;
}

StatusCode RichTimingScanAnalysis::book()
{
  m_RichSys = getDet<DeRichSystem>(DeRichLocations::RichSystem);

  const LHCb::RichSmartID::Vector &m_activeHPDs  = m_RichSys->allPDRichSmartIDs();
  info() << "number of currently active HPDs " << m_activeHPDs.size() << endmsg;
  Rich::DAQ::Level0ID l0ID;

  for (auto iHPD = m_activeHPDs.begin(); iHPD != m_activeHPDs.end(); ++iHPD)
  {
    std::ostringstream ostream;
    ostream << *iHPD;

    try
    {
      l0ID = m_RichSys->level0ID(*iHPD);
    }
    catch ( GaudiException g)
    {
      warning() << "Caught the exception: " << g.what() << endmsg;
      continue;
    }

    unsigned int RichNum = (unsigned int)iHPD->rich()+1;
    if (RichNum == m_whichRich)
    {
      info() << "L0ID " << l0ID.l0() << " HPD " << l0ID.hpd() << " HPDID " << m_RichSys->hardwareID(*iHPD) << " DATA " << l0ID.data();
      // histograms

      std::ostringstream hpdString,l0String;
      hpdString << "HPD_" << m_RichSys->hardwareID(*iHPD);
      l0String  << "L0_"  << l0ID.l0();
      info() << " " << hpdString.str() << " " << l0String.str() << endmsg;

      // 1D histogram: #hits inclusive per HPD:
      m_HPDtimingHists.insert(std::make_pair(m_RichSys->hardwareID(*iHPD), book1D(hpdString.str(), hpdString.str(), (-0.5*m_stepWidth)+m_stepStart, (m_nSteps+0.5)*m_stepWidth+m_stepStart, (m_nSteps+1))));

      // 1D histogram: #hits inclusive per L0 board:
      if (!histoExists(l0String.str()))
      {
        m_L0timingHists.insert(std::make_pair(l0ID.l0(), book1D(l0String.str(), l0String.str(),(-0.5*m_stepWidth)+m_stepStart, (m_nSteps+0.5)*m_stepWidth+m_stepStart, (m_nSteps+1))));
      }
    }
  }// for iHPD
  return StatusCode::SUCCESS;

}

StatusCode RichTimingScanAnalysis::initialize()
{
  StatusCode sc = Rich::HistoAlgBase::initialize();
  if ( sc.isFailure() ) return sc;
  acquireTool( "RichSmartIDDecoder", m_SmartIDDecoder, 0, true );
  acquireTool( "RichSmartIDTool", m_idTool,  0, true );
  acquireTool( "RichHPDPixelClusterFinder", m_clusterTool, this);
  sc = sc && book();
  m_evt=1;
  m_step=0;
    return sc;
  //return StatusCode::SUCCESS;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode RichTimingScanAnalysis::execute()
{
  debug() << "==> Execute" << endmsg;
  const auto & l1Map = m_SmartIDDecoder->allRichSmartIDs();
  StatusCode sc = StatusCode::SUCCESS;
  m_step = get<LHCb::ODIN>( LHCb::ODINLocation::Default )->calibrationStep();
  if (!m_filterBX || (m_filterBX && checkODINBxID()))
  {
   
    if (l1Map.size() <= 0)
    {
      m_EmptyEvents++;
      m_evt++;
      StatusCode ret = true;
      return ret;
    }

    for ( const auto & L1 : l1Map )
    {
      for ( const auto & ingress : L1.second )
      {
        for ( const auto & HPD : ingress.second.pdData() )
        {
          const auto &hpdInfo = HPD.second;
          const auto  &smartIDHPD  = hpdInfo.pdID();

          Rich::DAQ::Level0ID l0ID;
          try
          {
            l0ID = m_RichSys->level0ID(smartIDHPD);
          }
          catch (GaudiException g)
          {
            debug() << "Caught the exception: " << g.what() << endmsg;
            continue ;
          }
          unsigned int RichNum = (unsigned int)smartIDHPD.rich() + 1;
          if (RichNum==m_whichRich)
          {
            std::ostringstream hpdString;
            hpdString << "HPD_" << m_RichSys->hardwareID(smartIDHPD);
            if (histoExists(hpdString.str()))
            {
              const LHCb::RichSmartID::Vector  &smartIDHits = hpdInfo.smartIDs();  // SmartID vector of hits in HPD
              LHCb::RichSmartID::Vector hitMap(smartIDHits);
              const int nHits = smartIDHits.size();
              debug() << "L0ID " << l0ID.l0() << " HPDID " << m_RichSys->hardwareID(smartIDHPD) << " DATA " << l0ID.data();
              if(m_minClusSize>1 && m_maxClusSize < 999999)
              {
                auto clusters = m_clusterTool->findClusters(hitMap);
                const auto & clusVec = clusters->clusters();
                debug() << " Clusters found: " << clusVec.size();
                for (auto iClust = clusVec.begin(); iClust != clusVec.end(); ++iClust)      // loop over found clusters
                {
                  unsigned int clusterSize = (*(*iClust)).size();
                  if(clusterSize >= m_minClusSize && clusterSize < m_maxClusSize)
                  {
                    debug() << " Cluster size: " << clusterSize << endmsg;
                    m_HPDtimingHists[m_RichSys->hardwareID(smartIDHPD)]->fill((m_step*m_stepWidth)+m_stepStart,clusterSize);
                    m_L0timingHists[l0ID.l0()]->fill(m_step*m_stepWidth+m_stepStart,clusterSize);
                  }

                }
              }
              else
              {
                debug() << " nHits: " << nHits << endmsg;
                m_HPDtimingHists[m_RichSys->hardwareID(smartIDHPD)]->fill((m_step*m_stepWidth)+m_stepStart,nHits);
                m_L0timingHists[l0ID.l0()]->fill(m_step*m_stepWidth+m_stepStart,nHits);
              }
            }
            else
            {
              debug() << "L0ID " << l0ID.l0() << " HPDID " << m_RichSys->hardwareID(smartIDHPD) << " DATA " << l0ID.data() << " Was not initialised!" << endmsg;
            }
          }
        }
      }
    }
  }

  if (m_nEventsPerStep > 0 && (m_evt% m_nEventsPerStep)==0)
  {
    info() << "***Processed " << (m_evt+1)*100/(m_nEventsPerStep*m_nSteps) << " % of events" << endmsg;
  }
  m_evt++;
  return StatusCode::SUCCESS;
}

//=============================================================================
